# Copyright (c) 2012, Intel Corporation
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of Intel Corporation nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""SMBIOS/DMI module."""

import bits
from collections import namedtuple
import redirect
import struct
import uuid

class SMBIOS(namedtuple('SMBIOS',
                        ['AnchorString', 'Checksum',
                         'Length', 'MajorVersion',
                         'MinorVersion', 'MaxStructureSize',
                         'EntryPointRevision', 'FormattedArea',
                         'IntermediateAnchorString', 'IntermediateChecksum',
                         'StructureTableLength', 'StructureTableAddress',
                         'NumberStructures', 'BCDRevision'])):
    fmt_str = "<4sBBBBHB5s5sBHIHB"
    size = struct.calcsize(fmt_str)
    def __repr__(self):
        return "SMBIOS(AnchorString={AnchorString!r}, Checksum={Checksum:#x}, Length={Length}, MajorVersion={MajorVersion}, MinorVersion={MinorVersion}, MaxStructureSize={MaxStructureSize}, EntryPointRevision={EntryPointRevision}, FormattedArea={FormattedArea!r}, IntermediateAnchorString={IntermediateAnchorString!r}, IntermediateChecksum={IntermediateChecksum:#x}, StructureTableLength={StructureTableLength}, StructureTableAddress={StructureTableAddress:#x}, NumberStructures={NumberStructures}, BCDRevision={BCDRevision})".format(**self._asdict())

    def structures(self):
        offset = 0
        while offset != self.StructureTableLength:
            sm_struct = SMBIOSStructure._make(struct.unpack_from(SMBIOSStructure.fmt_str, self._structure_memory, offset))
            sm_struct.mem = bits.memory(self.StructureTableAddress + offset, sm_struct.Length)
            offset += sm_struct.Length
            strings = []
            while True:
                s = ""
                while self._structure_memory[offset] != "\x00":
                    s += self._structure_memory[offset]
                    offset += 1
                offset += 1
                if not s:
                    if not strings:
                        offset += 1
                    break
                strings.append(s)
            sm_struct.strings = strings
            yield sm_struct

class SMBIOSStructure(namedtuple('SMBIOSStructure', ['Type', 'Length', 'Handle'])):
    fmt_str = "<BBH"
    size = struct.calcsize(fmt_str)

    def getstr(self, i):
        if i == 0:
            return "(none)"
        if (i - 1) > len(self.strings):
            return "(error: string index out of range)"
        return self.strings[i - 1]

def parse():
    mem = bits.memory(0xF0000, 0x10000)
    signature = struct.unpack("<I", "_SM_")
    for offset in range(0, len(mem), 16):
        if struct.unpack_from("<I", mem, offset) == signature:
            break
    else:
        raise Exception("No SMBIOS table found")
    sm = SMBIOS._make(struct.unpack_from(SMBIOS.fmt_str, mem, offset))
    sm._structure_memory = bits.memory(sm.StructureTableAddress, sm.StructureTableLength)
    return sm

def log_smbios_info():
    try:
        sm = parse()
        output = {}
        def decode_0(sm_struct, sm):
            bi = BIOSInformation(sm_struct, sm)
            return 'BIOS vendor {!r} version {!r} date {!r}'.format(bi.sm_struct.getstr(bi.Vendor), bi.sm_struct.getstr(bi.Version), bi.sm_struct.getstr(bi.ReleaseDate))
        def decode_1(sm_struct, sm):
            si = SystemInformation(sm_struct, sm)
            return 'System manufacturer {!r} product {!r} version {!r}'.format(si.sm_struct.getstr(si.Manufacturer), si.sm_struct.getstr(si.ProductName), si.sm_struct.getstr(si.Version))
        known_types = { 0: decode_0, 1: decode_1 }
        for sm_struct in sm.structures():
            if sm_struct.Type in known_types:
                output.setdefault(sm_struct.Type, []).append(known_types[sm_struct.Type](sm_struct, sm))
                if len(output) == len(known_types):
                    break

        print >>redirect.log, "SMBIOS information:"
        for key in sorted(known_types.iterkeys()):
            for s in output.get(key, ["No structure of type {} found".format(key)]):
                print >>redirect.log, "{}: {}".format(key, s)
    except:
        print >>redirect.log, "Error parsing SMBIOS information:"
        import traceback
        traceback.print_exc(redirect.log)

def dump_raw():
    sm = parse()
    print "=== Raw SMBIOS Information ==="
    print sm
    for sm_struct in sm.structures():
        print
        print 'Type:{}  Address:{:#x}  Length:{}'.format(sm_struct.Type, bits.memory_addr(sm_struct.mem), len(sm_struct.mem))
        bits.dumpmem(sm_struct.mem, bits.memory_addr(sm_struct.mem))
        for n, s in enumerate(sm_struct.strings, 1):
            print 'string[{}] = {!r}'.format(n, s)

class BIOSInformation(object):
    def __init__(self, sm_struct, sm):
        self.sm_struct = sm_struct
        self.Type, self.Length, self.Handle, self.Vendor, self.Version, self.StartingAddressSegment, self.ReleaseDate, self.ROMSize, self.Characteristics = struct.unpack_from("<BBHBBHBBQ", sm_struct.mem)
        if (sm.MajorVersion, sm.MinorVersion) >= (2,4):
            characteristic_bytes = 2
            self.MajorRelease, self.MinorRelease, self.ECMajorRelease, self.ECMinorRelease = struct.unpack_from("<BBBB", sm_struct.mem, 0x14)
        else:
            characteristic_bytes = self.Length - 0x12
        self.CharacteristicsExtensionBytes = [ord(b) for b in sm_struct.mem[0x12:0x12+characteristic_bytes]]

    def display(self):
        print 'Type 0 BIOS Information'
        print 'Length: {}'.format(self.Length)
        print 'Handle: {}'.format(self.Handle)
        print 'Vendor: {!r}'.format(self.sm_struct.getstr(self.Vendor))
        print 'BIOS Version: {!r}'.format(self.sm_struct.getstr(self.Version))
        print 'BIOS Starting Address Segment: {:#x}'.format(self.StartingAddressSegment)
        print 'BIOS Release Date: {!r}'.format(self.sm_struct.getstr(self.ReleaseDate))
        print 'BIOS ROM Size: {}*64kB'.format(self.ROMSize+1)
        print 'BIOS Characteristics: {:#x}'.format(self.Characteristics)
        if self.CharacteristicsExtensionBytes:
            print 'BIOS Characteristics Extension Bytes:', " ".join("{:#x}".format(b) for b in self.CharacteristicsExtensionBytes)
        if hasattr(self, "MajorRelease"):
            print 'System BIOS Release: {}.{}'.format(self.MajorRelease, self.MinorRelease)
            print 'Embedded Controller Firmware Release: {}.{}'.format(self.ECMajorRelease, self.ECMinorRelease)

class SystemInformation(object):
    def __init__(self, sm_struct, sm):
        self.sm_struct = sm_struct
        self.Type, self.Length, self.Handle, self.Manufacturer, self.ProductName, self.Version, self.SerialNumber = struct.unpack_from("<BBHBBBB", sm_struct.mem)
        if self.Length >= 0x19:
            self.UUID = uuid.UUID(bytes_le=sm_struct.mem[0x8:0x18])
            self.WakeupType = struct.unpack_from("<B", sm_struct.mem, 0x18)[0]
        if self.Length >= 0x1B:
            self.SKUNumber, self.Family = struct.unpack_from("<BB", sm_struct.mem, 0x19)

    def display(self):
        print 'Type 1 System Information'
        print 'Length: {}'.format(self.Length)
        print 'Handle: {}'.format(self.Handle)
        print 'Manufacturer: {!r}'.format(self.sm_struct.getstr(self.Manufacturer))
        print 'Product Name: {!r}'.format(self.sm_struct.getstr(self.ProductName))
        print 'Version: {!r}'.format(self.Version)
        print 'Serial Number: {!r}'.format(self.sm_struct.getstr(self.SerialNumber))
        if hasattr(self, "UUID"):
            print 'UUID: {}'.format(self.UUID)
            print 'Wakeup Type: {:#x}'.format(self.WakeupType)
        if hasattr(self, "SKUNumber"):
            print 'SKU Number: {!r}'.format(self.SKUNumber)
            print 'Family: {!r}'.format(self.sm_struct.getstr(self.Family))

def dump():
    try:
        sm = parse()
        known_types = { 0: BIOSInformation, 1: SystemInformation }
        print "SMBIOS information:"
        for sm_struct in sm.structures():
            if sm_struct.Type in known_types:
                print
                known_types[sm_struct.Type](sm_struct, sm).display()
    except:
        print "Error parsing SMBIOS information:"
        import traceback
        traceback.print_exc()
